
from numpy import where, append, array, cumsum, concatenate

# functions to check persistence diagrams and convert diagrams between formats
def check_diagram(D):
    '''Checks for persistence diagrams.
    
    Internal method to verify that birth values are non-negative and less than death values.

    Parameters
    ----------
    `D` : numpy.ndarray
        The input diagram to be checked.
    
    Returns
    -------
    None
    '''
    # make sure that birth values are <= death values
    # if D has one entry
    if len(D.shape) != 1:
        pers = D[:,1] - D[:,0]
        birth = D[:,0]
    else:
        pers = D[1] - D[0]
        birth = D[0]
    if where(pers < 0)[0].shape[0] > 0:
        raise Exception('Birth values have to be <= death values.')
    # make sure that birth values are >= 0
    if where(birth < 0)[0].shape[0] > 0:
        raise Exception('Birth values have to be non-negative.')

def preprocess_diagram(D, inf_replace_val = None, ret = False):
    '''Verify the format of a persistence diagram and convert to a standard format.
    
    This function can verify a persistence diagram from the ripser, gph, flagser, gudhi or cechmate packages 
    and convert any such diagram into a list of numpy arrays if desired (largely an internal functionality but
    can be used in a standalone fashion).
    
    Parameters
    ----------
    `D` : any
        The persistence diagram to be verified. An exception will be raised if `D` is not a persistence
        diagram computed from one of the aforementioned packages.
    `inf_replace_val` : float or int, default `None`
        The value with which `inf` values should be replaced, if desired.
    `ret` : bool, default `False`
        Whether or not to return a processed diagram.
    
    Returns
    -------
    None or list of numpy.ndarray
        If `ret` is `True` and the diagram is verified then a list is returned. The i-th element of 
        the returned list is the array of i-dimensional topological features in the diagram.
    '''
    # error check inf_replace_val
    def check_val(D, inf_replace_val):
        if inf_replace_val != None:
            if not isinstance(inf_replace_val, type(2.0)) and not isinstance(inf_replace_val, type(2)):
                raise Exception('inf_replace_val must be a number.')
            if inf_replace_val <= 0:
                raise Exception('inf_replace_val must be positive.')
            max_death = max([d[d[:,1] != float('inf')].max() if len(d[d[:,1] != float('inf')]) > 0 else 0 for d in D])
            if inf_replace_val < max_death:
                raise Exception('inf_replace_val should be at least as large as any death value in the diagram.')
            # replace vals
            D[0][D[0][:,1] == float('inf'),1] = inf_replace_val
            return D
        else:
            return D
    error_message = 'Diagrams must be computed from either the ripser, gph, flagser, gudhi or cechmat libraries.'
    # first check if the diagram is from ripser, gph or flagser
    if isinstance(D, dict):
        if not 'dgms' in D.keys():
            raise Exception(error_message)
        if not isinstance(D['dgms'],list):
            raise Exception(error_message)
        if set([type(x) for x in D['dgms']]) != set([type(array([0,1]))]):
            raise Exception(error_message)
        if set([len(x.shape) for x in D['dgms']]) != set([2]):
            raise Exception(error_message)
        if set([x.shape[1] for x in D['dgms']]) != set([2]):
            raise Exception(error_message)
        # convert to list of numpy arrays
        D = D['dgms']
        # perform final numeric diagram checks
        lst = [check_diagram(d) for d in D]
        if ret == True:
            return check_val(D, inf_replace_val)
    # now check if diagram is from cechmate or gudhi
    elif isinstance(D, list):
        if set([type(x) for x in D]) == set([type((1,2))]):
            if set([len(x) for x in D]) != set([2]):
                raise Exception(error_message)
            if set([type(x[1]) for x in D]) != set([type((1,2))]):
                raise Exception(error_message)
            if set([len(x[1]) for x in D]) != set([2]):
                raise Exception(error_message)
            dims = [x[0] for x in D]
            if any(e < 0 for e in dims):
                raise Exception(error_message)
            if set([type(d) for d in dims]) != set([type(1)]):
                raise Exception(error_message)
            # convert to list of numpy arrays
            max_dim = max(dims)
            res = [array([0,0]).reshape((1,2)) for i in range(max_dim + 1)]
            for pt in D:
                res[pt[0]] = append(res[pt[0]], array([pt[1][0], pt[1][1]]).reshape((1,2)),axis = 0)
            res = [r[range(1,len(r)),:] for r in res]
            lst = [check_diagram(d) for d in res]
            if ret == True:
                return check_val(res, inf_replace_val)
        elif set([type(x) for x in D]) == set([type(array([0,1]))]):
            if set([len(x.shape) for x in D]) != set([2]):
                raise Exception(error_message)
            if set([x.shape[1] for x in D]) != set([2]):
                raise Exception(error_message)
            # no need to convert as this is the base format
            # final numeric diagram checks
            lst = [check_diagram(d) for d in D]
            if ret == True:
                return check_val(D, inf_replace_val)
        else:
            raise Exception(error_message)
    else:
        raise Exception(error_message)
    
def preprocess_diagram_groups_for_inference(diagram_groups):
    '''Additional internal preprocessing for permutation tests.
    
    Parameters
    ----------
    `diagram_groups` : list of lists
        Contains the persistence diagrams for the permutation test.
    
    Returns
    -------
    List of dicts and list of int
        Each element in the first list contains the diagram (key 'diagram') and its index in the total list (key 'ind').
        The second list contains the cumulative sum of group sizes, starting with 0.
    '''
    if not isinstance(diagram_groups, type([0,1])):
        raise Exception('diagram_groups must be a list.')
    if set([type(x) for x in diagram_groups]) != set([type([0,1])]):
        raise Exception('Each element of diagram_groups must be a list.')
    group_sizes = [len(g) for g in diagram_groups]
    csum_group_sizes = concatenate([array([0]), cumsum(group_sizes)])
    diagram_groups = [[{'diagram':preprocess_diagram(diagram_groups[g][i], ret = True), 'ind':csum_group_sizes[g] + i} for i in range(len(diagram_groups[g]))] for g in range(len(diagram_groups))]
    return diagram_groups, csum_group_sizes



# inference with persistence diagrams
from tdads.distance import *
from multiprocessing import cpu_count, Pool
from numpy import ones, ndarray, percentile, array, float64, equal
from numpy.testing import assert_almost_equal
from itertools import repeat, combinations
from random import sample, choices
from inspect import getfullargspec
import warnings

# permutation test
class perm_test:
    def __init__(self, iterations:int = 20, dims:list = [0], p:float = 2.0, q:float = 2.0, paired:bool = False, n_cores:int = cpu_count() - 1):
        '''Group difference testing for persistence diagrams. Based on the paper Robinson and Turner 2017, with
        extra functionality detailed in Abdallah 2021.
        
        Parameters
        ----------
        `iterations` : int, default 20
            The number of permutation iterations to carry out. The smallest p-value that can
            be returned by this test is 1/(`iterations` + 1).
        `dims` : list of int, default [0]
            The list of homological dimensions in which to carry out the test.
        `p` : float, default 2.0
            The power parameter for the Wasserstein metric.
        `q` : float, default 2.0
            The loss function exponential parameter, must be at least 1. See Robinson and Turner 2017
            for details.
        `paired` : bool, default `False`
            Whether or not there is a second-order pairing between each group, i.e. if only
            permutations which shuffle the groups first, second, third, etc elements should
            be used. See Abdallah 2021 for details.

        Attributes
        ----------
        `iterations` : int
            The input `iterations` parameter.
        `dims` : list of int
            The input `dims` parameter.
        `p` : float
            The input `p` parameter.
        `q` : float
            The input `q` parameter.
        `paired` : bool
            The input `paired` parameter.
        `distances` : list of `tdads.distance.distance` objects
            The distance functions in each desired dimension.
        `group_sizes` : `None` initially, then list of int once `test` is called
            The integer sizes of the diagram groups provided for testing.
        `dist_mats` : `None` initially, then a list of ndarray's
            A list, with one element for each dimension in `dim`, of partially computed distance matrices between all diagrams in the groups.
            Once two diagrams have there distances calculated in each dimension, the corresponding entries in all elements of `dist_mats` will be updated from -1 to the
            actual distance value in that dimension.
        
        Citations
        ---------
        Robinson T, Turner K (2017). "Hypothesis testing for topological data analysis." https://link.springer.com/article/10.1007/s41468-017-0008-7.

        Abdallah H et al. (2021). "Statistical Inference for Persistent Homology applied to fMRI." https://github.com/hassan-abdallah/Statistical_Inference_PH_fMRI/blob/main/Abdallah_et_al_Statistical_Inference_PH_fMRI.pdf.
        '''
        if isinstance(iterations, type(1)) == False:
            raise Exception('iterations must be an integer.')
        if iterations < 1:
            raise Exception('iterations must be at least 1.')
        self.iterations = iterations
        if not isinstance(dims, type([0,1])):
            raise Exception('dims must be a list.')
        if set([type(d) for d in dims]) != set([type(1)]):
            raise Exception('Each dimension in dims must be an integer.')
        if min(dims) < 0:
            raise Exception('Each dimension in dims must be non-negative.')
        self.dims = dims
        distances = [distance(dim = d, p = p) for d in dims]
        self.distances = distances
        if not isinstance(q,type(2)) and not isinstance(q,type(2.0)):
            raise Exception('q must be a number.')
        if q < 1:
            raise Exception('q must be at least 1.')
        self.q = q
        if isinstance(n_cores, type(1)) == False:
            raise Exception('n_cores must be an integer.')
        if n_cores < 1:
            raise Exception('n_cores must be at least 1.')
        self.group_sizes = None
        self.dist_mats = None
        if isinstance(paired, type(True)) == False:
            raise Exception('paired must be True or False.')
        self.paired = paired
    def __str__(self):
        '''Describe a permutation test procedure based on the number of permutation iterations
        and whether the groups are paired or unpaired.'''
        if self.paired == True:
            start_str = 'Paired permutation test with '
        else:
            start_str = 'Non-paired permutation test with '
        s = start_str + str(self.iterations) + ' iterations.'
        return s
    def compute_loss(self, diagram_groups):
        '''Internal method to compute the loss function from Robinson and Turner 2017.
        This function should not be called directly.'''
        combs = concatenate([concatenate([g*ones((int(len(diagram_groups[g])*(len(diagram_groups[g]) - 1)/2),1)), concatenate([array(x).reshape(1,2) for x in combinations(range(len(diagram_groups[g])), 2)])], axis = 1) for g in range(len(self.group_sizes))]).astype(int)
        def get_distance(dim_ind, combination): # make sure this updates!!
            v = self.dist_mats[dim_ind][diagram_groups[combination[0]][combination[1]]['ind'], diagram_groups[combination[0]][combination[2]]['ind']]
            if v == -1:
                v = self.distances[dim_ind].compute(diagram_groups[combination[0]][combination[1]]['diagram'], diagram_groups[combination[0]][combination[2]]['diagram'])
                self.dist_mats[dim_ind][diagram_groups[combination[0]][combination[1]]['ind'], diagram_groups[combination[0]][combination[2]]['ind']] = v
            return v
        statistics = []
        for d_i in range(len(self.dims)):
            d_tots = [get_distance(d_i, comb) for comb in combs]
            # with Pool(processes=self.n_cores) as pool:
            #     d_tots = pool.starmap(get_distance, zip(repeat(dim), combs)) # each row, need to do this per dim
            for i in range(len(combs)):
                self.dist_mats[d_i][diagram_groups[combs[i, 0]][combs[i, 1]]['ind'], diagram_groups[combs[i, 0]][combs[i, 2]]['ind']] = d_tots[i]
            statistics.append(sum([sum([d_tots[x]**self.q for x in range(len(combs)) if combs[x,0] == g])/(len(diagram_groups[g])*(len(diagram_groups[g]) - 1)) for g in range(len(diagram_groups))]))
        return statistics
    def test(self, diagram_groups):
        '''Run the permutation test.
        
        Parameters
        ----------
        `diagram_groups` : list of lists
            The groups of persistence diagrams to be analyzed.

        Returns
        -------
        Dict
            Keys are 'test_statistics' for the test statistic in each dimension, 
            'permvals' for the null distribution in each dimension and 'p_values' for the
            p-values in each dimension. For example, `output['p_values']['1']` would give the
            p-value for the second homological dimension in `self.dims`.
        
        Examples
        --------
        >>> # create two groups of persistence diagrams
        >>> from ripser import ripser
        >>> import numpy as np
        >>> data1 = np.random((100,2))
        >>> data2 = np.random((100,2))
        >>> D1 = ripser(data1)
        >>> D2 = ripser(data2)
        >>> group1 = [D1, D2]
        >>> group2 = [D1, D2]
        >>> # create perm test object in dimensions 0 and 1
        >>> from tdads.inference import permutation_test
        >>> pt = permutation_test(dims = [0, 1], n_cores = 2)
        >>> # run test
        >>> res = pt.test([g1, g2])
        >>> # get p-values
        >>> res['p_values']

        Citations
        ---------
        Robinson T, Turner K (2017). "Hypothesis testing for topological data analysis." https://link.springer.com/article/10.1007/s41468-017-0008-7.

        Abdallah H et al. (2021). "Statistical Inference for Persistent Homology applied to fMRI." https://github.com/hassan-abdallah/Statistical_Inference_PH_fMRI/blob/main/Abdallah_et_al_Statistical_Inference_PH_fMRI.pdf.
        '''
        # test the diagram_groups and preprocess
        diagram_groups, csum_group_sizes = preprocess_diagram_groups_for_inference(diagram_groups)
        # update group_sizes
        self.group_sizes = [len(g) for g in diagram_groups]
        # more checks
        if min(self.group_sizes) < 2:
            raise Exception('Each group of diagrams must have at least 2 diagrams.')
        if self.paired == True and len(set(self.group_sizes)) > 1:
            raise Exception('When paired is True each group of diagrams must have the same number of elements.')
        # set up to store distance calculations
        n = sum(self.group_sizes)
        self.dist_mats = [-1*ones((n, n)) for d in self.dims]
        # compute test statistics
        test_statistics = self.compute_loss(diagram_groups)
        # generate permutations
        perm_values = [[] for d in self.dims]
        # iterate
        for iteration in range(self.iterations):
            if not self.paired:
                permuted_groups = []
                samples = [i for i in range(n)]
                for i in range(len(diagram_groups)):
                    ss = sample(samples, len(diagram_groups[i]))
                    gs = [max([i for i in range(len(csum_group_sizes)) if csum_group_sizes[i] <= X]) if len([i for i in range(len(csum_group_sizes)) if csum_group_sizes[i] <= X]) > 0 else 0 for X in ss]
                    permuted_groups.append([diagram_groups[gs[j]][ss[j] - csum_group_sizes[gs[j]]] for j in range(len(ss))])
                    samples = [x for x in samples if x not in ss]
            else:
                perm = [sample(range(len(diagram_groups)), len(diagram_groups)) for x in range(self.group_sizes[0])]
                permuted_groups = [[diagram_groups[perm[p][i]][p] for p in range(len(perm))] for i in range(len(diagram_groups))]
            permuted_statistics = self.compute_loss(permuted_groups)
            for i in range(len(self.dims)):
                perm_values[i].append(permuted_statistics[i])
        # gather all data in results
        perm_values_ret = {}
        p_values_ret = {}
        test_statistics_ret = {}
        for i in range(len(self.dims)):
            perm_values_ret[str(self.dims[i])] = perm_values[i]
            test_statistics_ret[str(self.dims[i])] = test_statistics[i]
            p_values_ret[str(self.dims[i])] = (sum(perm_values[i] <= test_statistics[i]) + 1)/(self.iterations + 1)
        return {'permvals':perm_values_ret, 'test_statistics':test_statistics_ret, 'p_values':p_values_ret}
    
class diagram_bootstrap:
    def __init__(self, diag_fun, dims:list = [0], num_samples:int = 20, distance_mat:bool = False, alpha:float = 0.05):
        '''Compute confidence sets for (the topological features within) persistence diagrams.
        
        Parameters
        ----------
        `diag_fun` : function of two variables `X` and `thresh`
            The persistent homology algorithm (Vietoris Rips) for the dataset `X` up to 
            radius `thresh`. The maximum homological dimension should be the same as the
            maximum value of `dims`.
        `dims` : list of int, default [0]
            The list of homological dimensions in which to compute confidence sets.
        `num_samples` : int, default 20
            The number of bootstrap resamplings to carry out.
        `distance_mat` : bool, default False
            Whether the input dataset will be a distance matrix or not.
        `alpha` : float, default 0.05
            The type 1 error for determining significant topological features.

        Attributes
        ----------
        `diag_fun` : function
            The input `diag_fun` parameter.
        `dims` : list
            The input `dims` parameter.
        `num_samples` : int
            The input `num_samples` parameter.
        `distance_mat` : bool, default False
            The input `distance_mat` parameter.
        `alpha` : float
            The input `alpha` parameter.

        Examples
        --------
        >>> from tdads.inference import diagram_bootstrap
        >>> from ripser import ripser
        >>> from numpy.random import uniform
        >>> # build circle dataset
        >>> theta = uniform(low = 0, high = 2*pi, size = 100)
        >>> data = array([[cos(theta[i]), sin(theta[i])] for i in range(100)])
        >>> # define persistent homology function
        >>> def diag_fun(X, thresh):
        >>>     return ripser(X = X, thresh = thresh)
        >>> # create bootstrap object
        >>> boot = diagram_bootstrap(diag_fun = diag_fun)
        
        Citations
        ---------
        Chazal F et al (2017). "Robust Topological Inference: Distance to a Measure and Kernel Distance." https://www.jmlr.org/papers/volume18/15-484/15-484.pdf.
        '''
        def check_fun(a, b):
            return 1
        if not isinstance(diag_fun, type(check_fun)):
            raise Exception('diag_fun must be a function.')
        if getfullargspec(diag_fun)[0] != ['X', 'thresh']:
            raise Exception('diag_fun must be a function of two parameters, X and thresh.')
        if not isinstance(dims, type([0,1])):
            raise Exception('dims must be a list.')
        if set([type(d) for d in dims]) != set([type(1)]):
            raise Exception('Each dimension in dims must be an integer.')
        if min(dims) < 0:
            raise Exception('Each dimension in dims must be non-negative.')
        self.dims = dims

        if isinstance(distance_mat, type(True)) == False:
            raise Exception('distance_mat must be True or False.')
        self.distance_mat = distance_mat

        diamond = array([[0,0],[1,1],[-1,1],[0,2]])
        dist_diamond = array([[0, sqrt(2), sqrt(2), 2], [sqrt(2), 0, 2, sqrt(2)], [sqrt(2), 2, 0, sqrt(2)], [2, sqrt(2), sqrt(2), 0]])
        diamond_diag = [array([[0, sqrt(2)],[0, sqrt(2)],[0, sqrt(2)],[0, float('inf')]]),array([[sqrt(2), 2]])]
        diamond_diag = [diamond_diag[x] if x < len(diamond_diag) else array([]).reshape(0,2).astype(float64) for x in self.dims]
        if not distance_mat:
            X = diamond
        else:
            X = dist_diamond
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                sample_res = diag_fun(X, float('inf'))
            sample_res = preprocess_diagram(sample_res, ret = True)
        except Exception as ex:
            raise Exception('diag_fun doesn\'t seem to be computing persistence diagrams correctly.')
        for i in self.dims:
            try:
                assert_almost_equal(sample_res[i], diamond_diag[self.dims.index(i)], 7)
            except:
                raise Exception('diag_fun doesn\'t seem to be computing persistence diagrams correctly.')
        self.diag_fun = diag_fun

        if isinstance(num_samples, type(1)) == False:
            raise Exception('num_samples must be an integer.')
        if num_samples < 1:
            raise Exception('num_samples must be at least 1.')
        self.num_samples = num_samples

        if not isinstance(alpha,type(0.05)):
            raise Exception('alpha must be a float.')
        if alpha <= 0 or alpha >= 1:
            raise Exception('alpha must be between 0 and 1 (non-inclusive).')
        self.alpha = alpha
    def __str__(self):
        '''Describe a bootstrap procedure based on the number of bootstrap samples,
        whether or not the input will be a distance matrix and the Type 1 error rate (alpha).'''
        dms = ''
        if not self.distance_mat:
            dms = 'non-'
        s = 'Bootstrap confidence intervals with ' + str(self.num_samples) + ' many samples, ' + '' +  'distance-matrix input, and a Type 1 error of ' + str(self.alpha) + '.'
        return s
    def compute(self, X:ndarray, thresh:float):
        '''Carry out the bootstrap procedure.
        
        Parameters
        ----------
        `X` : numpy.ndarray (2D)
            The input dataset - either raw tabular data or a distance matrix of samples.
        `thresh` : float
            The maximum filtration radius for Vietoris-Rips persistent homology.

        Returns
        -------
        Dict
            Entries are 'diagram' (the computed persistence diagram), 'thresholds' (a Dict of the computed
            persistence thresholds for each desired dimension) and 'subsetted_diagram' (the persistence diagram
            thresholded by the threshold values in each dimension).

        Examples
        --------
        >>> from tdads.inference import diagram_bootstrap
        >>> from ripser import ripser
        >>> from numpy.random import uniform
        >>> # build circle dataset
        >>> theta = uniform(low = 0, high = 2*pi, size = 100)
        >>> data = array([[cos(theta[i]), sin(theta[i])] for i in range(100)])
        >>> # define persistent homology function
        >>> def diag_fun(X, thresh):
        >>>     return ripser(X = X, thresh = thresh)
        >>> # create bootstrap object and compute significant features
        >>> boot = diagram_bootstrap(diag_fun = diag_fun)
        >>> res = boot.compute(data, 2)
        >>> # print subsetted diagram
        >>> res['subsetted_diagram']

        Citations
        ---------
        Chazal F et al (2017). "Robust Topological Inference: Distance to a Measure and Kernel Distance." https://www.jmlr.org/papers/volume18/15-484/15-484.pdf.
        '''
        if not isinstance(X, type(array([0,1]))):
            raise Exception('X must be a numpy array.')
        if len(X.shape) != 2 or X.shape[0] < 2 or X.shape[1] < 1:
            raise Exception('X must be a 2-dimensional array with at least two rows and one column.')
        if self.distance_mat and X.shape[0] != X.shape[1]:
            raise Exception('When distance_mat is True X must have the same number of rows and columns (as a distance matrix).')
        if not (isinstance(thresh, type(1)) or isinstance(thresh, type(0.1))):
            raise Exception('thresh must be a number.')
        if thresh <= 0:
            raise Exception('thresh must be positive.')
        # try to calculate the full persistence diagram
        try:
            diagram = self.diag_fun(X, thresh)
        except Exception as ex:
            raise Exception('An error occured when diag_fun tried to compute the persistence diagram of X. The error was: ' + str(ex))
        # error check the persistence diagram
        try:
            diagram = preprocess_diagram(D = diagram, ret = True)
        except Exception as ex:
            raise Exception('The output of diagam_fun(X, thresh) was not in the correct format for a persistence diagram.')
        # create bottleneck distance objects in each dimension
        distances = [distance(dim = d, p = float('inf'), n_cores = 2) for d in self.dims]
        # function to bootstrap
        def get_distances():
            # generate sample (unique row indices)
            s = choices(population = range(len(X)), k = len(X))
            s = list(set(s))
            # subset X
            if self.distance_mat:
                X_sample = X[s,:][:,s]
            else:
                X_sample = X[s, :]
            # compute new persistence diagram
            try:
                diagram_sample = self.diag_fun(X_sample, thresh)
            except Exception as ex:
                raise Exception('An output of diag_fun was not in the correct format for a persistence diagram.')
            # compute distances in each desired dimension
            res = [dist.compute(diagram, diagram_sample) for dist in distances]
            return res
        # perform bootstrap
        bootstrap_values = [get_distances() for i in range(self.num_samples)]
        # filter by dimension
        bootstrap_values = [array([bv[i] for bv in bootstrap_values]) for i in range(len(self.dims))]
        # compute thresholds
        thresholds = [2*percentile(bv, 1 - self.alpha) for bv in bootstrap_values]
        # subset diagram by thresholds
        subsetted_diagram = [diagram[i][diagram[i][:,1] - diagram[i][:,0] > thresholds[self.dims.index(i)]] if i in self.dims else empty((0, 2)) for i in range(max(self.dims) + 1)]
        # set up return dict
        ret = {'diagram':diagram, 'thresholds':thresholds, 'subsetted_diagram':subsetted_diagram}
        return ret

        
    



# functions to calculate distances between diagrams
from tdads.diagram_utils import *
from scipy.optimize import linear_sum_assignment
from scipy.spatial.distance import cdist
from numpy import concatenate, array, logical_or, exp, sqrt, arccos, empty, zeros
from math import pi
from multiprocessing import Pool, cpu_count
from itertools import product

# add in extra parameter for n_cores in distance constructor
class distance:
    def __init__(self, dim:int = 0,metric='W',p:float=2.0, sigma:float=None, inf_replace_val:float = None, n_cores:int=cpu_count() - 1):
        '''Distance metrics for persistence diagrams.
        
        Available distance metrics are the Wasserstein, bottleneck and Fisher information metric distances.

        Parameters
        ----------
        `dim` : int, default 0
            The non-negative homological dimension in which distances will be computed .
        `metric` : str, default \"W\"
            One of \"W\" or \"FIM\" for the Wasserstein/bottleneck and Fisher information metric
            functions respectively.
        `p` : float, default 2.0
            The power parameter for the Wasserstein metric, must be at least 1.
        `sigma` : float, default `None`
            The scale parameter for the Fisher information metric, must be supplied when
            `metric` is \"FIM\".
        `inf_replace_val` : float or int, default `None`
            The value with which `inf` values should be replaced, if desired. If `None`, topological features with `inf` values will 
            be ignored, otherwise original diagrams will be modified.
        `n_cores` : int, default is the number of available cores minus 1.
            The number of CPU cores to use for parallel computation of distance matrices.
        
        Attributes
        ----------
        `dim` : int
            The input `dim` parameter.
        `metric` : str
            The input `metric` parameter.
        `p` : float
            The input `p` parameter.
        `sigma` : float
            The input `sigma` parameter.
        `inf_replace_val` : float
            The input `inf_replace_val` parameter.
        `n_cores` : int
            The input `n_cores` parameter.
        '''
        if not isinstance(dim,type(2)):
            raise Exception('dim must be an integer.')
        if dim < 0:
            raise Exception('dim must be non-negative.')
        self.dim = dim
        if not isinstance(metric,type('')):
            raise Exception('The distance metric must be a string.')
        if metric != 'W' and metric != 'FIM':
            raise Exception('The distance metric must be either \"W\" for the wasserstein/bottleneck distances or \"FIM\" for the Fisher information metric.')
        self.metric = metric
        if metric == 'W':
            if not isinstance(p,type(2)) and not isinstance(p,type(2.0)):
                raise Exception('For the wasserstein/bottleneck distance p must be a number.')
            if p < 1:
                raise Exception('For the wasserstein/bottleneck distance p must be at least 1.')
            self.p = p
            self.sigma = None
        else:
            if not isinstance(sigma,type(2)) and not isinstance(sigma,type(2.0)):
                raise Exception('For the Fisher information metric sigma must be a number.')
            if sigma <= 0:
                raise Exception('For the Fisher information metric sigma must be positive.')
            self.sigma = sigma
            self.p = None
        if inf_replace_val != None:
            if not isinstance(inf_replace_val, type(2.0)) and not isinstance(inf_replace_val, type(2)):
                raise Exception('inf_replace_val must be a number.')
            if inf_replace_val <= 0:
                raise Exception('inf_replace_val must be positive.')
        self.inf_replace_val = inf_replace_val
        if not isinstance(n_cores,type(2)):
            raise Exception('n_cores must be an integer.')
        if n_cores < 0:
            raise Exception('n_cores must be non-negative.')
        if n_cores > cpu_count():
            raise Exception('n_cores must be at most the number of available cores.')
        # then check if n_cores is more than the number of available cores
        self.n_cores = n_cores
    def __str__(self):
        '''Describe a distance metric by type (Wasserstein, bottleneck or 
        Fisher information metric) and major parameter (`p` for Wasserstein and 
        `sigma` for Fisher information metric).'''
        if self.sigma != None:
            M = 'Fisher information metric with parameter sigma = ' + str(self.sigma) + '.'
        else:
            if self.p == float('inf'):
                M = 'Bottleneck distance.'
            else:
                M = str(self.p) + '-wasserstein distance.'
        return M
    def compute(self,D1,D2) -> float:
        '''Compute the distance between two persistence diagrams.

        Parameters
        ----------
        `D1` : any
            The first persistence diagram (computed from either the ripser, gph, flagser, gudhi or cechmate packages).
        `D2` : any
            The second persistence diagram (\"\").
        
        Returns
        -------
        float
            The numeric distance calculation value. In dimension 0 persistence diagrams may contain a point whose death is inf,
            and these points are ignored (if you wish to use these points make sure to replace inf with the maximum filtration value
            you used to compute the diagram).

        Examples
        --------
        >>> from tdads import distance
        >>> from ripser import ripser
        >>> import numpy as np
        >>> # create 2 datasets
        >>> data1 = np.random((100,2))
        >>> data2 = np.random((100,2))
        >>> # compute persistence diagrams with ripser
        >>> diagram1 = ripser(data1)
        >>> diagram2 = ripser(data2)
        >>> # create distance object
        >>> d_wass = distance() # 2-wasserstein distance
        >>> # compute distance
        >>> d_wass.compute(diagram1, diagrams2)

        Citations
        ---------
        Kerber M, Morozov D and Nigmetov A (2017). "Geometry Helps to Compare Persistence Diagrams." https://dl.acm.org/doi/10.1145/3064175.
        
        Le T, Yamada M (2018). "Persistence fisher kernel: a riemannian manifold kernel for persistence diagrams." https://proceedings.neurips.cc/paper/2018/file/959ab9a0695c467e7caf75431a872e5c-Paper.pdf.
        
        Vlad I. Morariu, Balaji Vasan Srinivasan, Vikas C. Raykar, Ramani Duraiswami, and Larry S. Davis. Automatic online tuning for fast Gaussian summation. Advances in Neural Information Processing Systems (NIPS), 2008.
        '''
        # preprocess diagrams
        D1 = preprocess_diagram(D1, ret=True, inf_replace_val = self.inf_replace_val)
        D2 = preprocess_diagram(D2, ret=True, inf_replace_val = self.inf_replace_val)
        # subset diagrams to correct dimensions
        if len(D1) > self.dim:
            D1_sub = D1[self.dim]
        else:
            D1_sub = empty((0,2))
        if len(D2) > self.dim:
            D2_sub = D2[self.dim]
        else:
            D2_sub = empty((0,2))
        # remove infinity's in diagrams
        D1_sub = D1_sub[D1_sub[:,1] != float('inf')]
        D2_sub = D2_sub[D2_sub[:,1] != float('inf')]
        # remove diagonal points from both diagrams
        D1_sub = D1_sub[(D1_sub[:,0] < D1_sub[:,1]),:]
        D2_sub = D2_sub[(D2_sub[:,0] < D2_sub[:,1]),:]
        # check for empty diagrams, return distance 0
        if D1_sub.shape[0] == 0 and D2_sub.shape[0] == 0:
            return 0
        # compute diagonal projections and concatenate to opposite diagram
        if D1_sub.shape[0] > 0:
            diag1 = concatenate([0.5*array([x[0]+x[1], x[0]+x[1]]).reshape((1,2)) for x in D1_sub])
            n_diag1 = diag1.shape[0]
        else:
            diag1 = empty((0,2))
            n_diag1 = 0
        if D2_sub.shape[0] > 0:
            diag2 = concatenate([0.5*array([x[0]+x[1], x[0]+x[1]]).reshape((1,2)) for x in D2_sub])
            n_diag2 = diag2.shape[0]
        else:
            diag2 = empty((0,2))
            n_diag2 = 0
        D2_sub = concatenate([D2_sub, diag1])
        D1_sub = concatenate([D1_sub, diag2])
        # compute distance
        if self.metric == 'W':
            # wasserstein/bottleneck distance
            # calculate max (chebyshev) distance between each pair of rows in D1_sub and D2_sub
            dist_mat = cdist(D1_sub, D2_sub, metric='chebyshev')
            # set distances between points and their projections to 0
            v1 = D1_sub.shape[0] - n_diag2
            v2 = D2_sub.shape[0] - n_diag1
            if v1 < D1_sub.shape[0] and v2 < D2_sub.shape[0]:
                prod = concatenate([array(x).reshape((1,2)) for x in product(range(v1, D1_sub.shape[0]),range(v2, D2_sub.shape[0]))])
                dist_mat[prod[:,0], prod[:,1]] = 0
            # if wasserstein distance then exponentiate
            if self.p < float('inf'):
                dist_mat = dist_mat**self.p
            # solve linear sum assignment problem
            row_ind, col_ind = linear_sum_assignment(cost_matrix = dist_mat, maximize = False)
            # subset to remove matches between projection points
            good_inds = logical_or(row_ind < v1, col_ind < v2)
            row_ind = row_ind[good_inds]
            col_ind = col_ind[good_inds]
            # return distance value
            if self.p < float('inf'):
                return dist_mat[row_ind, col_ind].sum()**(1/self.p)
            else:
                return dist_mat[row_ind, col_ind].max()
        else:
            # Fisher information metric distance
            theta = concatenate([D1_sub, D2_sub])
            # exact calculation for now...
            rho1 = [exp(-1*cdist(x.reshape((1,2)), D1_sub, metric='euclidean')**2/(2*self.sigma**2)).sum()/(sqrt(2*pi)*self.sigma) for x in theta]
            rho2 = [exp(-1*cdist(x.reshape((1,2)), D2_sub, metric='euclidean')**2/(2*self.sigma**2)).sum()/(sqrt(2*pi)*self.sigma) for x in theta]
            # check for same vectors
            if rho1 == rho2:
                return 0
            # normalize
            rho1 = rho1/sum(rho1)
            rho2 = rho2/sum(rho2)
            # dot product
            norm = sum([sqrt(x)*sqrt(y) for x,y in zip(rho1, rho2)])
            # check bounds and return arc cos
            if norm > 1:
                norm = 1
            if norm < -1:
                norm = -1
            return arccos(norm)
    def compute_matrix(self, diagrams:list, other_diagrams:list=None):
        '''Compute a distance matrix between one or two lists of persistence diagrams.
        Parameters
        ----------
        `diagrams` : list
            The first first of persistence diagram (computed from either the ripser, gph, flagser, gudhi or cechmate packages).
        `other_diagrams` : any
            The optional second list of persistence diagram for computing a cross-distance matrix. Default `None`.
        
        Returns
        -------
        numpy.ndarray
            The (cross) distance matrix.

        Examples
        --------
        >>> from tdads import distance
        >>> from ripser import ripser
        >>> import numpy as np
        >>> # create 2 datasets
        >>> data1 = np.random((100,2))
        >>> data2 = np.random((100,2))
        >>> # compute persistence diagrams with ripser
        >>> diagram1 = ripser(data1)
        >>> diagram2 = ripser(data2)
        >>> # create distance object
        >>> d_wass = distance() # 2-wasserstein distance
        >>> # compute distance matrix
        >>> d_wass.compute_matrix([diagram1, diagram2])
        >>> # this is the same as:
        >>> d_wass.compute_matrix([diagram1, diagram2], [diagram1, diagram2])
        '''
        # first check the lists of diagrams
        # then create a joint list of pairs of diagrams for parallel processing
        if isinstance(diagrams,type([1,2])) == False:
            raise Exception('diagrams must be a list.')
        if len(diagrams) == 0:
            raise Exception('diagrams list must not be empty.')
        diagrams = [preprocess_diagram(d, ret = True) for d in diagrams]
        if other_diagrams != None:
            if isinstance(other_diagrams, type([1,2])) == False:
                raise Exception('If supplied, other_diagrams must be a list.')
            if len(other_diagrams) == 0:
                raise Exception('If supplied, other_diagrams list must not be empty.')
            other_diagrams = [preprocess_diagram(d, ret = True) for d in other_diagrams]
            prod = [x for x in product(range(len(diagrams)), range(len(other_diagrams)))]
            diagram_product = product(diagrams, other_diagrams)
            D = zeros((len(diagrams), len(other_diagrams)))
        else:
            diagram_product = product(diagrams, diagrams)
            prod = [x for x in product(range(len(diagrams)), range(len(diagrams)))]
            prod = [x for x in prod if x[0] < x[1]]
            diagram_product = iter([(diagrams[p[0]], diagrams[p[1]]) for p in prod])
            D = zeros((len(diagrams), len(diagrams)))
        # create a pool process and compute distances in parallel
        with Pool(processes=self.n_cores) as pool:
            result = pool.starmap(self.compute, diagram_product)
        # store result in numpy array
        for i in range(len(prod)):
            p = prod[i]
            D[p[0], p[1]] = result[i]
            if other_diagrams == None:
                D[p[1], p[0]] = result[i]
        return D
        
        

        
        



# read version from installed package
from importlib.metadata import version
__version__ = version("tdads")

# only include public functions here
from .diagram_utils import check_diagram, preprocess_diagram
from .distance import distance
from .kernel import kernel
from .machine_learning import *
from .inference import *

__all__ = [
    'check_diagram',
    'preprocess_diagram',
    'distance',
    'kernel',
    'diagram_mds',
    'diagram_kpca',
    'perm_test',
    'diagram_bootstrap'
]



# machine learning of persistence diagrams
from tdads.distance import *
from tdads.kernel import *
from multiprocessing import cpu_count
from sklearn.manifold import MDS
from sklearn.decomposition import KernelPCA
from sklearn.svm import SVC, SVR
from numpy import concatenate, array

# multidimensional scaling
class diagram_mds:
    '''Multidimensional scaling with persistence diagrams.'''
    def __init__(self, n_components:int = 2, random_state:int = None, precomputed:bool = False, dim:int = 0, metric:str = 'W', p:float = 2, sigma:float = None, n_cores:int = cpu_count() - 1):
        '''Multidimensional scaling with persistence diagrams.
        
        Parameters
        ----------
        `n_components` : int
            The number of dimensions in which to disperse the distance values, default 2.
        `random_state` : int
            Determines the random number generator used for reproducibility, default None.
        `precomputed` : bool
            Determines whether a precomputed distance matrix of persistence diagrams (`True`) or a list of persistence diagrams (`False`, the default) will be passed to the fit method.
        `dim` : int
            The non-negative homological dimension in which distances will be computed (default 0).
        `metric` : str
            One of \"W\" (default) or \"FIM\" for the wasserstein/bottleneck and Fisher information metric
            functions respectively.
        `p` : float
            The power parameter for the wasserstein metric, must be at least 1 (default 2).
        `sigma` : float
            The scale parameter for the Fisher information metric, default None but must be supplied when
            `metric` is \"FIM\".
        `n_cores` : int
            The number of CPU cores to use for parallel computation of distance matrices. Default is the
            number of available cores minus 1.

        Attributes
        ----------
        `distance` : tdads.distance.distance
            The object used to compute the distance matrix of persistence diagrams. 
        `MDS` : sklearn.manifold._mds.MDS
            The sklearn.manifold.MDS object used for embedding the distance matrix.
        `precomputed` : bool
            The input `precomputed` parameter.
        '''
        self.distance = distance(dim = dim, metric = metric, p = p, sigma = sigma, n_cores = n_cores)
        self.MDS = MDS(n_components = n_components, metric = False, n_jobs = n_cores, random_state = random_state, dissimilarity = 'precomputed')
        if isinstance(precomputed, type(True)) == False:
            raise Exception('precomputed must be True or False.')
        self.precomputed = precomputed
    def __str__(self):
        '''Describe a persistence diagram multidimensional scaling object via its distance metric.'''
        s = 'Non-metric multidimensional scaling of persistence diagrams. Distance metric used: ' + self.distance.__str__
        return s
    def fit_transform(self, X, y:any = None):
        '''Fit the data in X and compute the position of the persistence diagrams in the embedding space.
        
        Parameters
        ----------
        `X` : {array-like of shape `(n_diagrams, n_diagrams)`} or {list of length `n_diagrams`}
            Either a precomputed distance matrix of `n_diagrams` many persistence diagrams (if `precomputed` was set to `True`) or a list of `n_diagrams` many persistence diagrams (otherwise).
        `y` : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        `X_new` : ndarray of shape `(n_diagrams, n_components)`
            `X` transformed in the new space.

        Examples
        --------
        >>> from tdads.machine_learning import diagram_mds
        >>> from tdads.distance import distance
        >>> from ripser import ripser
        >>> import numpy as np
        >>> # create 2 datasets
        >>> data1 = np.random((100,2))
        >>> data2 = np.random((100,2))
        >>> # compute persistence diagrams with ripser
        >>> diagram1 = ripser(data1)
        >>> diagram2 = ripser(data2)
        >>> # project into 2D with the 2-wasserstein distance
        >>> mds = diagram_mds()
        >>> mds.fit_transform([D1, D2])
        >>> # can also fit with a precomputed distance matrix
        >>> d_wass = distance()
        >>> dist_mat = d_wass.compute_matrix([D1, D2])
        >>> mds_precomp = diagram_mds(precomputed = True)
        >>> mds_precomp.fit_transform(dist_mat)
        '''
        if self.precomputed == False:
            if isinstance(X, type(array([0,1]))):
                raise Exception('When precomputed is False, X must be a list of persistence diagrams.')
            X = self.distance.compute_matrix(X)
        else:
            if isinstance(X, type([0,1])):
                raise Exception('When precomputed is True, X must be a ndarray distance matrix of persistence diagrams.')
        X_new = self.MDS.fit_transform(X, y)
        return X_new
    
# kernel PCA
class diagram_kpca:
    '''Kernel PCA with persistence diagrams.'''
    def __init__(self, n_components:int = 2, random_state:int = None, precomputed:bool = False, diagrams:list = None, dim:int = 0, sigma:float = 1.0, t:float = 1.0, n_cores:int = cpu_count() - 1):
        '''Multidimensional scaling with persistence diagrams.
        
        Parameters
        ----------
        `n_components` : int
            The number of dimensions in which to disperse the kernel values, default 2.
        `random_state` : int
            Determines the random number generator used for reproducibility, default None.
        `precomputed` : bool
            Determines whether a precomputed Gram matrix of persistence diagrams (`True`) or a list of persistence diagrams (`False`, the default) will be passed to the fit method.
        `diagrams` : list
            An optional list of persistence diagrams. If `precomputed` is `False` then `diagrams` will take the value of the
            list of diagrams passed to the `fit` method. Otherwise, `diagrams` must be supplied in order to perform inference (i.e. prediction).
        `dim` : int
            The non-negative homological dimension in which distances will be computed (default 0).
        `metric` : str
            One of \"W\" (default) or \"FIM\" for the wasserstein/bottleneck and Fisher information metric
            functions respectively.
        `sigma` : float
            The scale parameter for the Fisher information metric, default 1.
        `t` : float
            The positive dispersion parameter for the persistence Fisher kernel, default 1.
        `n_cores` : int
            The number of CPU cores to use for parallel computation of distance matrices. Default is the
            number of available cores minus 1.
        
        Attributes
        ----------
        `kernel` : tdads.kernel.kernel
            The object used to compute the (cross) Gram matrices of persistence diagrams. 
        `kPCA` : sklearn.decomposition._kernel_pca.KernelPCA
            The kernel PCA object used for embedding the persistence diagrams.
        `precomputed` : bool
            The input `precomputed` parameter.
        `diagrams` : list of length `n_diagrams`
            The input `diagrams` parameter for inference.
        '''
        self.kernel = kernel(dim = dim, sigma = sigma, t = t, n_cores = n_cores)
        self.kPCA = KernelPCA(n_components = n_components, n_jobs = n_cores, random_state = random_state, kernel = 'precomputed')
        if isinstance(precomputed, type(True)) == False:
            raise Exception('precomputed must be True or False.')
        self.precomputed = precomputed
        if isinstance(diagrams, type([1,2])) == False:
            raise Exception('diagrams must be a list of diagrams.')
        self.diagrams = diagrams
    def __str__(self):
        '''Describe a persistence diagram kernel principle components analysis object via its kernel function.'''
        s = 'Kernel PCA of persistence diagrams. Kernel used: ' + self.kernel.__str__
        return s
    def fit(self, X, y:any = None):
        '''Fit the model from data in X.
        
        Parameters
        ----------
        `X` : {array-like of shape `(n_diagrams, n_diagrams)`} or {list of length `n_diagrams`}
            Either a precomputed Gram matrix of `n_diagrams` many persistence diagrams (if `precomputed` was set to `True`) or a list of `n_diagrams` many persistence diagrams (otherwise).
        `y` : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        `self` : object
            Returns the instance itself.

        Examples
        --------
        >>> from tdads.machine_learning import diagram_mds
        >>> from tdads.kernel import kernel
        >>> from ripser import ripser
        >>> import numpy as np
        >>> # create 2 datasets
        >>> data1 = np.random((100,2))
        >>> data2 = np.random((100,2))
        >>> # compute persistence diagrams with ripser
        >>> diagram1 = ripser(data1)
        >>> diagram2 = ripser(data2)
        >>> # fit model with the persistence Fisher kernel (sigma = t = 1)
        >>> kpca = diagram_kpca()
        >>> kpca_fitted = kpca.fit([D1, D2])
        >>> # can also fit with a precomputed distance matrix
        >>> pfk = kernel()
        >>> gram_mat = pfk.compute_matrix([D1, D2])
        >>> kpca_precomp = diagram_kpca(precomputed = True)
        >>> kpca_precomp_fitted = kpca_precomp.fit(gram_mat)
        '''
        if self.precomputed == False:
            if isinstance(X, type(array([0,1]))):
                raise Exception('When precomputed is False, X must be a list of persistence diagrams.')
            self.diagrams = X
            X = self.kernel.compute_matrix(X)
        else:
            if isinstance(X, type([0,1])):
                raise Exception('When precomputed is True, X must be a ndarray distance matrix of persistence diagrams.')
        self.kPCA = self.kPCA.fit(X, y)
        return self
    def transform(self, X):
        '''Project new persistence diagrams into the embedding space.
        
        Parameters
        ----------
        `X` : {array-like of shape `(n_diagrams, n_diagrams)`} or {list of length `n_diagrams`}
            Either a precomputed (cross) Gram matrix of shape `(n_new_diagrams, n_diagrams)` (between the new persistence diagrams and the 
            training set diagrams, if `precomputed` was set to `True`) or a list of `n_new_diagrams` many persistence diagrams (otherwise).

        Returns
        -------
        `X_new` : ndarray
            The embedding of the new persistence diagrams.
        
        Examples
        --------
        >>> from tdads.machine_learning import diagram_mds
        >>> from tdads.kernel import kernel
        >>> from ripser import ripser
        >>> import numpy as np
        >>> # create 2 datasets
        >>> data1 = np.random((100,2))
        >>> data2 = np.random((100,2))
        >>> # compute persistence diagrams with ripser
        >>> diagram1 = ripser(data1)
        >>> diagram2 = ripser(data2)
        >>> # fit models (regular and precomputed) with the 
        >>> # persistence Fisher kernel (sigma = t = 1)
        >>> kpca = diagram_kpca()
        >>> kpca_fitted = kpca.fit([D1, D2]) # or
        >>> pfk = kernel()
        >>> gram_mat = pfk.compute_matrix([D1, D2])
        >>> kpca_precomp = diagram_kpca(precomputed = True)
        >>> kpca_precomp_fitted = kpca_precomp.fit(gram_mat)
        >>> # create 2 new datasets
        >>> data3 = np.random((100,2))
        >>> data4 = np.random((100,2))
        >>> # project new data into 2D space
        >>> kpca_fitted.transform([D3, D4]) # or
        >>> cross_gram = pfk.compute_matrix([D1, D2], [D3, D4])
        >>> kpca_precomputed_fitted.transform([D3, D4])
        '''
        if self.precomputed == False:
            X = self.kernel.compute_matrix(X, self.diagrams)
        return self.kPCA.transform(X)
    def fit_transform(self, X, y:any = None):
        '''Fit the data in X and compute the position of the persistence diagrams in the embedding space.
        
        Parameters
        ----------
        `X` : {array-like of shape `(n_diagrams, n_diagrams)`} or {list of length `n_diagrams`}
            Either a precomputed Gram matrix of `n_diagrams` many persistence diagrams (if `precomputed` was set to `True`) or a list of `n_diagrams` many persistence diagrams (otherwise).
        `y` : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        `X_new` : ndarray
            `X` transformed in the new space.
        
        Examples
        --------
        >>> from tdads.machine_learning import diagram_mds
        >>> from tdads.kernel import kernel
        >>> from ripser import ripser
        >>> import numpy as np
        >>> # create 2 datasets
        >>> data1 = np.random((100,2))
        >>> data2 = np.random((100,2))
        >>> # compute persistence diagrams with ripser
        >>> diagram1 = ripser(data1)
        >>> diagram2 = ripser(data2)
        >>> # fit models (regular and precomputed) with the 
        >>> # persistence Fisher kernel (sigma = t = 1) and
        >>> # project into 2D space
        >>> kpca = diagram_kpca()
        >>> kpca.fit_transform([D1, D2]) # or
        >>> pfk = kernel()
        >>> gram_mat = pfk.compute_matrix([D1, D2])
        >>> kpca_precomp = diagram_kpca(precomputed = True)
        >>> kpca_precomp.fit_transform(gram_mat)
        '''
        self = self.fit(X, y)
        X_new = self.transform(X)
        return X_new

# class diagram_svm():
#     '''Support vector machine for persistence diagrams.'''
#     def __init__(self, diagrams:list = None, cv:int = 1, dims:list = [0], sigmas:list = [1.0], ts:list = [1.0], precomputed:bool = False, Cs:list = [1.0], epsilons:list = [0.1], n_cores:int = cpu_count() - 1):
#         '''Support vector machines for persistence diagrams.
        
#         Parameters
#         ----------
#         `diagrams` : list of persistence diagrams, default None
#             When `precomputed = True`, `diagrams` must be supplied in order to call the predict method.
#         `cv` : int, default 1
#             The number of folds for cross validattion. The default is no cross-validation.
#         `dims` : list of int, default [0]
#             The homological dimensions in which to fit SVM models.
#         `sigmas` : list of float, default [1.0]
#             The values of `sigma` for the persistence Fisher kernel.
#         `ts` : list of float, default [1.0]
#             The values of `t` for the persistence Fisher kernel.
#         `precomputed` : bool, default False
#             If `True` then the `fit` method will expect precomputed Gram matrices for training, otherwise
#             a list of persistence diagrams.
#         `Cs` : list of float, default [1.0]
#             A list of regularization parameters. The strength of the regularization is inversely proportional to C. 
#             Must be strictly positive. The penalty is a squared l2.
#         `epsilons` : list of float, default [0.1]
#             A list of epsilons in the epsilon-SVR model. If performing classification set `epsilon = None`. 
#             `epsilons` specifies the epsilon-tubes within which no penalty is associated 
#             in the training loss function with points predicted within a distance epsilon from the actual value. 
#             Must be non-negative.
#         `n_cores` : int
#             The number of CPU cores to use for parallel computation of distance matrices. Default is the
#             number of available cores minus 1.

#         Attributes
#         ----------
#         `precomputed` : bool
#             The input `precomputed` parameter.
#         `cv` : int
#             The input `cv` parameter. 
#         `diagrams` : list of length `n_diagrams`
#             The input `diagrams` parameter for inference when `precomputed` is `False`.
#         `n_cores` : int
#             The input `n_cores` parameter.
#         `parameter_grid` : ndarray either of shape `(num_param_combos, 5)` or `(num_param_combos, 4)`
#             The cartesian product of all possible model parameter combinations (the number of which is `num_param_combos`).
#             The columns give the values of `dims`, `sigmas`, `ts`, `Cs` and `epsilons` (for regression) in that order, resulting
#             in five columns for classification and four for regression.
#         `models` : list of sklearn.svm._classes.SVR or sklearn.svm._classes.SVC of length `num_param_combos`
#             One model for each row of `parameter_grid` (to be fit with those parameters).
#         `final_model` : None or {sklearn.svm._classes.SVR or sklearn.svm._classes.SVC}
#             Initially None but becomes the optimal model object based on cross-validation results
#             once the `fit` method has been called.
#         `final_model_kernel` : None or tdads.kernel.kernel
#             Initially None but becomes the kernel object with parameters determined by `final_model`
#             once the `fit` method has been called.
#         '''
#         if isinstance(precomputed, type(True)) == False:
#             raise Exception('precomputed must be True or False.')
#         self.precomputed = precomputed
#         if isinstance(cv, type(1)) == False:
#             raise Exception('cv must be an integer.')
#         if cv < 1:
#             raise Exception('cv must be at least 1.')
#         self.cv = cv
#         if isinstance(n_cores, type(1)) == False:
#             raise Exception('n_cores must be an integer.')
#         if n_cores < 1:
#             raise Exception('n_cores must be at least 1.')
#         self.n_cores = n_cores
#         if set([type(d) for d in dims]) != set([type(1)]):
#             raise Exception('Each dimension in dims must be an integer.')
#         if min(dims) < 0:
#             raise Exception('Each dimension in dims must be non-negative.')
#         if set([x in set([type(1), type(1.0)]) for x in set([type(s) for s in sigmas])]) != set([True]):
#             raise Exception('Each sigma value must be a number.')
#         if min(sigmas) <= 0:
#             raise Exception('Each sigma value must be positive.')
#         if set([x in set([type(1), type(1.0)]) for x in set([type(t) for t in ts])]) != set([True]):
#             raise Exception('Each t value must be a number.')
#         if min(ts) <= 0:
#             raise Exception('Each t value must be positive.')
#         if isinstance(Cs, type([0,1])) == False:
#             raise Exception('Cs must be a list.')
#         if isinstance(epsilons, type([0,1])) == False:
#             raise Exception('epsilons must be a list.')
#         if epsilons != None:
#             parameter_grid = product(dims, sigmas, ts, Cs, epsilons)
#         else:
#             parameter_grid = product(dims, sigmas, ts, Cs)
#         parameter_grid = concatenate([[array(x)] for x in parameter_grid])
#         if epsilons == None:
#             self.models = [SVC(C=parameter_grid[i,3]) for i in range(len(parameter_grid))]
#         else:
#             self.models = [SVR(C=parameter_grid[i,3],epsilon=parameter_grid[i,4]) for i in range(len(parameter_grid))]
#         self.final_model = None
#         self.parameter_grid = parameter_grid
#         self.final_model_kernel = None
#     def __str__(self):
#         if self.parameter_grid.shape[1] == 4:
#             task = 'classification'
#         else:
#             task = 'regression'
#         if self.final_model == None:
#             fit_str = 'Model has not yet been fit.'
#         else:
#             fit_str = 'Model has been fit.'
#         return 'Support vector ' + task + ' object. ' + fit_str
#     def fit(self, X, y):
#         '''Fit the SVM model according to the training data.
        
#         Parameters
#         ----------
#         `X` : {array-like of shape `(n_diagrams, n_diagrams)`} or {list of length `n_diagrams`}
#             Either a precomputed Gram matrix of `n_diagrams` many persistence diagrams (if `precomputed` was set to `True`) or a list of `n_diagrams` many persistence diagrams (otherwise).
#         `y` : array-like of shape `(n_diagrams,)`
#             Target values (class labels in classification, real numbers in regression).

#         Returns
#         -------
#         `self` : object
#             The fitted estimator.

#         Examples
#         --------
#         # DO!
#         '''
#         # make row memberships for cv
#         if self.parameter_grid.shape[1] == 5:
#             1
#         else:
#             1





# kernel methods for persistence diagrams
# currently just the persistence Fisher kernel
from tdads.diagram_utils import *
from tdads.distance import *
from multiprocessing import cpu_count
from numpy import exp

class kernel:
    def __init__(self, dim:int = 0, sigma:float = 1, t:float = 1, inf_replace_val:float = None, n_cores:int = cpu_count() - 1):
        '''Kernel functions for persistence diagrams.
        
        Current available option is the persistence Fisher kernel.

        Parameters
        ----------
        `dim` : int, default 0
            The non-negative homological dimension in which kernels will be computed.
        `sigma` : float, default 1
            The scale parameter for the Fisher information metric.
        `t` : float, default 1
            The positive dispersion parameter for the persistence Fisher kernel.
        `inf_replace_val` : float or int, default `None`
            The value with which `inf` values should be replaced, if desired. If `None`, topological features with `inf` values will 
            be ignored, otherwise original diagrams will be modified.
        `n_cores` : int, default is the number of available cores minus 1
            The number of CPU cores to use for parallel computation of distance matrices.
        
        Attributes
        ----------
        `distance` : `tdads.distance.distance`
            The Fisher information metric distance object.
        `t` : float
            The input `t` parameter.
        '''
        self.dist = distance(dim = dim, metric = 'FIM', sigma = sigma, inf_replace_val = inf_replace_val, n_cores = n_cores)
        if not isinstance(t,type(2)) and not isinstance(t,type(2.0)):
            raise Exception('For the persistence Fisher kernel, t must be a number.')
        if t <= 0:
            raise Exception('For the persistence Fisher kernel, t must be positive.')
        self.t = t
    def __str__(self):
        '''Describe a persistence Fisher kernel by its `sigma` and `t` parameters.'''
        return 'Persistence Fisher kernel with sigma = ' + str(self.dist.sigma) + ', t = ' + str(self.t) + '.'
    def compute(self, D1, D2):
        '''Compute the kernel value between two persistence diagrams.

        Parameters
        ----------
        `D1` : any
            The first persistence diagram (computed from either the ripser, gph, flagser, gudhi or cechmate packages).
        `D2` : any
            The second persistence diagram.
        
        Returns
        -------
        float
            The numeric kernel calculation value.

        Examples
        --------
        >>> from tdads import kernel
        >>> from ripser import ripser
        >>> import numpy as np
        >>> # create 2 datasets
        >>> data1 = np.random((100,2))
        >>> data2 = np.random((100,2))
        >>> # compute persistence diagrams with ripser
        >>> diagram1 = ripser(data1)
        >>> diagram2 = ripser(data2)
        >>> # create kernel object
        >>> k = kernel()
        >>> # compute kernel value
        >>> k.compute(diagram1, diagrams2)

        Citations
        ---------        
        Le T, Yamada M (2018). "Persistence fisher kernel: a riemannian manifold kernel for persistence diagrams." https://proceedings.neurips.cc/paper/2018/file/959ab9a0695c467e7caf75431a872e5c-Paper.pdf.
        '''
        d_FIM = self.dist.compute(D1, D2)
        return exp(-1*self.t*d_FIM)
    def compute_matrix(self, diagrams, other_diagrams = None):
        '''Compute a Gram (kernel) matrix between one or two lists of persistence diagrams.
        Parameters
        ----------
        `diagrams` : list
            The first first of persistence diagram (computed from either the ripser, gph, flagser, gudhi or cechmate packages).
        `other_diagrams` : any
            The optional second list of persistence diagram for computing a cross-Gram matrix. Default `None`.
        
        Returns
        -------
        numpy.ndarray
            The (cross) Gram matrix.

        Examples
        --------
        >>> from tdads import kernel
        >>> from ripser import ripser
        >>> import numpy as np
        >>> # create 2 datasets
        >>> data1 = np.random((100,2))
        >>> data2 = np.random((100,2))
        >>> # compute persistence diagrams with ripser
        >>> diagram1 = ripser(data1)
        >>> diagram2 = ripser(data2)
        >>> # create kernel object
        >>> k = kernel()
        >>> # compute Gram matrix
        >>> k.compute_matrix([diagram1, diagram2])
        >>> # this is the same as:
        >>> k.compute_matrix([diagram1, diagram2], [diagram1, diagram2])
        '''
        dm_FIM = self.dist.compute_matrix(diagrams, other_diagrams)
        return exp(-1*self.t*dm_FIM)